@@ -21,7 +21,11 @@
 #include "lvm-string.h"
 #include "fs.h"
 #include "defaults.h"
+#include "segtypes.h"
+#include "display.h"
 #include "toolcontext.h"
+#include "targets.h"
+#include "config.h"
 
 #include <libdevmapper.h>
 #include <limits.h>
@@ -62,13 +66,8 @@ enum {
 	SUSPENDED = 4,
 	NOPROPAGATE = 5,
 	TOPLEVEL = 6,
-	REMOVE = 7
-};
-
-enum {
-	MIRR_DISABLED,
-	MIRR_RUNNING,
-	MIRR_COMPLETED
+	REMOVE = 7,
+	RESUME_IMMEDIATE = 8
 };
 
 typedef enum {
@@ -112,14 +111,14 @@ struct dl_list {
 };
 
 static const char *stripe_filler = NULL;
-static uint32_t mirror_region_size = 0;
 
 struct dev_manager {
 	struct pool *mem;
 
-	struct config_tree *cft;
+	struct cmd_context *cmd;
+
 	const char *stripe_filler;
-	uint32_t mirror_region_size;
+	void *target_state;
 	uint32_t pvmove_mirror_count;
 
 	char *vg_name;
@@ -166,89 +165,6 @@ static inline void _clear_flag(struct dev_layer *dl, int bit)
 	dl->flags &= ~(1 << bit);
 }
 
-/*
- * Device layer names are all of the form <vg>-<lv>-<layer>, any
- * other hyphens that appear in these names are quoted with yet
- * another hyphen.  The top layer of any device has no layer
- * name.  eg, vg0-lvol0.
- */
-static void _count_hyphens(const char *str, size_t *len, int *hyphens)
-{
-	const char *ptr;
-
-	for (ptr = str; *ptr; ptr++, (*len)++)
-		if (*ptr == '-')
-			(*hyphens)++;
-}
-
-/*
- * Copies a string, quoting hyphens with hyphens.
- */
-static void _quote_hyphens(char **out, const char *src)
-{
-	while (*src) {
-		if (*src == '-')
-			*(*out)++ = '-';
-
-		*(*out)++ = *src++;
-	}
-}
-
-/*
- * <vg>-<lv>-<layer> or if !layer just <vg>-<lv>.
- */
-static char *_build_name(struct pool *mem, const char *vg,
-			 const char *lv, const char *layer)
-{
-	size_t len = 0;
-	int hyphens = 0;
-	char *r, *out;
-
-	_count_hyphens(vg, &len, &hyphens);
-	_count_hyphens(lv, &len, &hyphens);
-
-	if (layer && *layer)
-		_count_hyphens(layer, &len, &hyphens);
-
-	len += hyphens + 2;
-
-	if (!(r = pool_alloc(mem, len))) {
-		stack;
-		return NULL;
-	}
-
-	out = r;
-	_quote_hyphens(&out, vg);
-	*out++ = '-';
-	_quote_hyphens(&out, lv);
-
-	if (layer && *layer) {
-		*out++ = '-';
-		_quote_hyphens(&out, layer);
-	}
-	*out = '\0';
-
-	return r;
-}
-
-/* Find start of LV component in hyphenated name */
-static char *_find_lv_name(char *vg)
-{
-	char *c = vg;
-
-	while (*c && *(c + 1)) {
-		if (*c == '-') {
-			if (*(c + 1) == '-')
-				c++;
-			else
-				return (c + 1);
-		}
-		c++;
-	}
-
-	return NULL;
-}
-
 static char *_build_dlid(struct pool *mem, const char *lvid, const char *layer)
 {
 	char *dlid;
@@ -336,7 +252,7 @@ static int _info_run(const char *name, const char *uuid, struct dm_info *info,
 static int _info(const char *name, const char *uuid, int mknodes,
 		 struct dm_info *info, struct pool *mem, char **uuid_out)
 {
-	if (!mknodes && uuid && *uuid && 
+	if (!mknodes && uuid && *uuid &&
 	    _info_run(NULL, uuid, info, 0, mem, uuid_out) && info->exists)
 		return 1;
 
@@ -427,18 +343,16 @@ static int _percent_run(struct dev_manager *dm, const char *name,
 	uint64_t start, length;
 	char *type = NULL;
 	char *params = NULL;
-	float percent2;
 	struct list *segh = &lv->segments;
 	struct lv_segment *seg = NULL;
+	struct segment_type *segtype;
 
-	uint64_t numerator, denominator;
 	uint64_t total_numerator = 0, total_denominator = 0;
 
 	*percent = -1;
 
 	if (!(dmt = _setup_task(name, uuid, event_nr,
-				wait ? DM_DEVICE_WAITEVENT : DM_DEVICE_STATUS)))
-	{
+				wait ? DM_DEVICE_WAITEVENT : DM_DEVICE_STATUS))) {
 		stack;
 		return 0;
 	}
@@ -471,40 +385,19 @@ static int _percent_run(struct dev_manager *dm, const char *name,
 		if (!type || !params || strcmp(type, target_type))
 			continue;
 
-		/* Mirror? */
-		if (!strcmp(type, "mirror")) {
-			log_debug("Mirror status: %s", params);
-			if (sscanf(params, "%*d %*x:%*x %*x:%*x %" PRIu64
-				   "/%" PRIu64, &numerator,
-				   &denominator) != 2) {
-				log_error("Failure parsing mirror status: %s",
-					  params);
-				goto out;
-			}
-			total_numerator += numerator;
-			total_denominator += denominator;
-
-			if (seg && (seg->status & PVMOVE))
-				seg->extents_moved = dm->mirror_region_size *
-				    numerator / lv->vg->extent_size;
+		if (!(segtype = get_segtype_from_string(dm->cmd, type)))
 			continue;
+
+		if (segtype->ops->target_percent &&
+		    !segtype->ops->target_percent(&dm->target_state, dm->mem,
+						  dm->cmd->cft, seg, params,
+						  &total_numerator,
+						  &total_denominator,
+						  percent)) {
+			stack;
+			goto out;
 		}
 
-		if (strcmp(type, "snapshot"))
-			continue;
-
-		/* Snapshot */
-		if (index(params, '/')) {
-			if (sscanf(params, "%" PRIu64 "/%" PRIu64,
-				   &numerator, &denominator) == 2) {
-				total_numerator += numerator;
-				total_denominator += denominator;
-			}
-			continue;
-		} else if (sscanf(params, "%f", &percent2) == 1) {
-			*percent += percent2;
-			*percent /= 2;
-		}
 	} while (next);
 
 	if (lv && (segh = list_next(&lv->segments, segh))) {
@@ -515,7 +408,7 @@ static int _percent_run(struct dev_manager *dm, const char *name,
 
 	if (total_denominator)
 		*percent = (float) total_numerator *100 / total_denominator;
-	else
+	else if (*percent < 0)
 		*percent = 100;
 
 	log_debug("LV percent: %f", *percent);
@@ -543,10 +436,16 @@ static int _percent(struct dev_manager *dm, const char *name, const char *uuid,
 	return 0;
 }
 
-static int _rename(struct dev_layer *dl, char *newname)
+static int _rename(struct dev_manager *dm, struct dev_layer *dl, char *newname)
 {
 	int r = 1;
 	struct dm_task *dmt;
+	char *vgname, *lvname, *layer;
+
+	if (!split_dm_name(dm->mem, dl->name, &vgname, &lvname, &layer)) {
+		log_error("Couldn't split up dm layer name %s", dl->name);
+		return 0;
+	}
 
 	log_verbose("Renaming %s to %s", dl->name, newname);
 
@@ -561,11 +460,13 @@ static int _rename(struct dev_layer *dl, char *newname)
 		goto out;
 	}
 
-	if (!(r = dm_task_run(dmt)))
+	if (!(r = dm_task_run(dmt))) {
 		log_error("Couldn't rename device '%s'.", dl->name);
+		goto out;
+	}
 
 	if (r && _get_flag(dl, VISIBLE))
-		fs_rename_lv(dl->lv, newname, _find_lv_name(dl->name));
+		fs_rename_lv(dl->lv, newname, lvname);
 
 	dl->name = newname;
 
@@ -574,6 +475,55 @@ static int _rename(struct dev_layer *dl, char *newname)
 	return r;
 }
 
+static int _suspend_or_resume(const char *name, action_t suspend)
+{
+	int r;
+	struct dm_task *dmt;
+	int sus = (suspend == SUSPEND) ? 1 : 0;
+	int task = sus ? DM_DEVICE_SUSPEND : DM_DEVICE_RESUME;
+
+	log_very_verbose("%s %s", sus ? "Suspending" : "Resuming", name);
+	if (!(dmt = _setup_task(name, NULL, 0, task))) {
+		stack;
+		return 0;
+	}
+
+	if (!(r = dm_task_run(dmt)))
+		log_error("Couldn't %s device '%s'", sus ? "suspend" : "resume",
+			  name);
+
+	dm_task_destroy(dmt);
+	return r;
+}
+
+static int _suspend(struct dev_layer *dl)
+{
+	if (!dl->info.exists || dl->info.suspended)
+		return 1;
+
+	if (!_suspend_or_resume(dl->name, SUSPEND)) {
+		stack;
+		return 0;
+	}
+
+	dl->info.suspended = 1;
+	return 1;
+}
+
+static int _resume(struct dev_layer *dl)
+{
+	if (!dl->info.exists || !dl->info.suspended)
+		return 1;
+
+	if (!_suspend_or_resume(dl->name, RESUME)) {
+		stack;
+		return 0;
+	}
+
+	dl->info.suspended = 0;
+	return 1;
+}
+
 static int _load(struct dev_manager *dm, struct dev_layer *dl, int task)
 {
 	int r = 1;
@@ -633,9 +583,9 @@ static int _load(struct dev_manager *dm, struct dev_layer *dl, int task)
 		log_error("Couldn't load device '%s'.", dl->name);
 		if ((dl->lv->minor >= 0 || dl->lv->major >= 0) &&
 		    _get_flag(dl, VISIBLE))
-			log_error("Perhaps the persistent device number "
-				  "%d:%d is already in use?",
-				  dl->lv->major, dl->lv->minor);
+			    log_error("Perhaps the persistent device number "
+				      "%d:%d is already in use?",
+				      dl->lv->major, dl->lv->minor);
 	}
 
 	if (!dm_task_get_info(dmt, &dl->info)) {
@@ -650,6 +600,13 @@ static int _load(struct dev_manager *dm, struct dev_layer *dl, int task)
 		goto out;
 	}
 
+	if (_get_flag(dl, RESUME_IMMEDIATE) && dl->info.suspended &&
+	    !_resume(dl)) {
+		stack;
+		r = 0;
+		goto out;
+	}
+
 	log_very_verbose("Activated %s %s %03u:%03u", dl->name,
 			 dl->dlid, dl->info.major, dl->info.minor);
 
@@ -696,55 +653,6 @@ static int _remove(struct dev_layer *dl)
 	return r;
 }
 
-static int _suspend_or_resume(const char *name, action_t suspend)
-{
-	int r;
-	struct dm_task *dmt;
-	int sus = (suspend == SUSPEND) ? 1 : 0;
-	int task = sus ? DM_DEVICE_SUSPEND : DM_DEVICE_RESUME;
-
-	log_very_verbose("%s %s", sus ? "Suspending" : "Resuming", name);
-	if (!(dmt = _setup_task(name, NULL, 0, task))) {
-		stack;
-		return 0;
-	}
-
-	if (!(r = dm_task_run(dmt)))
-		log_error("Couldn't %s device '%s'", sus ? "suspend" : "resume",
-			  name);
-
-	dm_task_destroy(dmt);
-	return r;
-}
-
-static int _suspend(struct dev_layer *dl)
-{
-	if (!dl->info.exists || dl->info.suspended)
-		return 1;
-
-	if (!_suspend_or_resume(dl->name, SUSPEND)) {
-		stack;
-		return 0;
-	}
-
-	dl->info.suspended = 1;
-	return 1;
-}
-
-static int _resume(struct dev_layer *dl)
-{
-	if (!dl->info.exists || !dl->info.suspended)
-		return 1;
-
-	if (!_suspend_or_resume(dl->name, RESUME)) {
-		stack;
-		return 0;
-	}
-
-	dl->info.suspended = 0;
-	return 1;
-}
-
 /*
  * The functions that populate the table in a dm_task as part of
  * a create/reload.
@@ -759,98 +667,26 @@ static int _emit_target_line(struct dev_manager *dm, struct dm_task *dmt,
 			     size_t paramsize)
 {
 	uint64_t esize = seg->lv->vg->extent_size;
-	uint32_t s, start_area = 0u, areas = seg->area_count;
-	int w = 0, tw = 0;
+	int w = 0;
 	const char *target = NULL;
-	const char *trailing_space;
-	int mirror_status;
-	struct dev_layer *dl;
-	char devbuf[10];
+	int r;
 
-	switch (seg->type) {
-	case SEG_SNAPSHOT:
+	if (!seg->segtype->ops->compose_target_line) {
 		log_error("_emit_target: Internal error: Can't handle "
-			  "SEG_SNAPSHOT");
+			  "segment type %s", seg->segtype->name);
 		return 0;
-		/* Target formats:
-		 *   linear [device offset]+
-		 *   striped #stripes stripe_size [device offset]+
-		 *   mirror  log_type #log_params [log_params]* 
-		 *           #mirrors [device offset]+
-		 */
-	case SEG_STRIPED:
-		if (areas == 1)
-			target = "linear";
-		else if (areas > 1) {
-			target = "striped";
-			if ((tw = lvm_snprintf(params, paramsize, "%u %u ",
-					       areas, seg->stripe_size)) < 0)
-				goto error;
-			w = tw;
-		} else {
-			log_error("_emit_target: Internal error: SEG_STRIPED "
-				  "with no stripes");
-			return 0;
-		}
-		break;
-	case SEG_MIRRORED:
-		mirror_status = MIRR_RUNNING;
-		if (seg->status & PVMOVE) {
-			if (seg->extents_moved == seg->area_len) {
-				mirror_status = MIRR_COMPLETED;
-				start_area = 1;
-			} else if (dm->pvmove_mirror_count++) {
-				mirror_status = MIRR_DISABLED;
-				areas = 1;
-			}
-		}
-		if (mirror_status != MIRR_RUNNING) {
-			target = "linear";
-			break;
-		}
-		target = "mirror";
-		if ((tw = lvm_snprintf(params, paramsize, "core 1 %u %u ",
-				       dm->mirror_region_size, areas)) < 0)
-			goto error;
-		w = tw;
-		break;
 	}
 
-	for (s = start_area; s < areas; s++, w += tw) {
-		trailing_space = (areas - s - 1) ? " " : "";
-		if ((seg->area[s].type == AREA_PV &&
-		     (!seg->area[s].u.pv.pv || !seg->area[s].u.pv.pv->dev)) ||
-		    (seg->area[s].type == AREA_LV && !seg->area[s].u.lv.lv))
-			tw = lvm_snprintf(params + w, paramsize - w,
-					  "%s 0%s", dm->stripe_filler,
-					  trailing_space);
-		else if (seg->area[s].type == AREA_PV)
-			tw = lvm_snprintf(params + w, paramsize - w,
-					  "%s %" PRIu64 "%s",
-					  dev_name(seg->area[s].u.pv.pv->dev),
-					  (seg->area[s].u.pv.pv->pe_start +
-					   (esize * seg->area[s].u.pv.pe)),
-					  trailing_space);
-		else {
-			if (!(dl = hash_lookup(dm->layers,
-					       seg->area[s].u.lv.lv->lvid.s))) {
-				log_error("device layer %s missing from hash",
-					  seg->area[s].u.lv.lv->lvid.s);
-				return 0;
-			}
-			if (!dm_format_dev(devbuf, sizeof(devbuf), dl->info.major, dl->info.minor)) {
-				log_error("Failed to format device number as dm target (%u,%u)",
-					  dl->info.major, dl->info.minor);
-				return 0;
-			}
-			tw = lvm_snprintf(params + w, paramsize - w,
-					  "%s %" PRIu64 "%s", devbuf,
-					  esize * seg->area[s].u.lv.le,
-					  trailing_space);
-		}
-
-		if (tw < 0)
-			goto error;
+	if ((r = seg->segtype->ops->compose_target_line(dm, dm->mem,
+							dm->cmd->cft,
+							&dm->target_state, seg,
+							params, paramsize,
+							&target, &w,
+							&dm->
+							pvmove_mirror_count)) <=
+	    0) {
+		stack;
+		return r;
 	}
 
 	log_debug("Adding target: %" PRIu64 " %" PRIu64 " %s %s",
@@ -863,11 +699,62 @@ static int _emit_target_line(struct dev_manager *dm, struct dm_task *dmt,
 	}
 
 	return 1;
+}
 
-      error:
-	log_debug("Insufficient space in params[%" PRIsize_t "] for target "
-		  "parameters.", paramsize);
-	return -1;
+int compose_areas_line(struct dev_manager *dm, struct lv_segment *seg,
+		       char *params, size_t paramsize, int *pos, int start_area,
+		       int areas)
+{
+	uint32_t s;
+	int tw = 0;
+	const char *trailing_space;
+	uint64_t esize = seg->lv->vg->extent_size;
+	struct dev_layer *dl;
+	char devbuf[10];
+
+	for (s = start_area; s < areas; s++, *pos += tw) {
+		trailing_space = (areas - s - 1) ? " " : "";
+		if ((seg->area[s].type == AREA_PV &&
+		     (!seg->area[s].u.pv.pv || !seg->area[s].u.pv.pv->dev)) ||
+		    (seg->area[s].type == AREA_LV && !seg->area[s].u.lv.lv))
+			tw = lvm_snprintf(params + *pos, paramsize - *pos,
+					  "%s 0%s", dm->stripe_filler,
+					  trailing_space);
+		else if (seg->area[s].type == AREA_PV)
+			tw = lvm_snprintf(params + *pos, paramsize - *pos,
+					  "%s %" PRIu64 "%s",
+					  dev_name(seg->area[s].u.pv.pv->dev),
+					  (seg->area[s].u.pv.pv->pe_start +
+					   (esize * seg->area[s].u.pv.pe)),
+					  trailing_space);
+		else {
+			if (!(dl = hash_lookup(dm->layers,
+					       seg->area[s].u.lv.lv->lvid.s))) {
+				log_error("device layer %s missing from hash",
+					  seg->area[s].u.lv.lv->lvid.s);
+				return 0;
+			}
+			if (!dm_format_dev
+			    (devbuf, sizeof(devbuf), dl->info.major,
+			     dl->info.minor)) {
+				log_error
+				    ("Failed to format device number as dm target (%u,%u)",
+				     dl->info.major, dl->info.minor);
+				return 0;
+			}
+			tw = lvm_snprintf(params + *pos, paramsize - *pos,
+					  "%s %" PRIu64 "%s", devbuf,
+					  esize * seg->area[s].u.lv.le,
+					  trailing_space);
+		}
+
+		if (tw < 0) {
+			stack;
+			return -1;
+		}
+	}
+
+	return 1;
 }
 
 static int _emit_target(struct dev_manager *dm, struct dm_task *dmt,
@@ -892,6 +779,9 @@ static int _emit_target(struct dev_manager *dm, struct dm_task *dmt,
 		if (ret >= 0)
 			return ret;
 
+		log_debug("Insufficient space in params[%" PRIsize_t
+			  "] for target parameters.", paramsize);
+
 		paramsize *= 2;
 	} while (paramsize < MAX_TARGET_PARAMSIZE);
 
@@ -1023,8 +913,8 @@ static int _populate_snapshot(struct dev_manager *dm,
 /*
  * dev_manager implementation.
  */
-struct dev_manager *dev_manager_create(const char *vg_name,
-				       struct config_tree *cft)
+struct dev_manager *dev_manager_create(struct cmd_context *cmd,
+				       const char *vg_name)
 {
 	struct pool *mem;
 	struct dev_manager *dm;
@@ -1039,22 +929,16 @@ struct dev_manager *dev_manager_create(const char *vg_name,
 		goto bad;
 	}
 
+	dm->cmd = cmd;
 	dm->mem = mem;
-	dm->cft = cft;
+
 	if (!stripe_filler) {
-		stripe_filler = find_config_str(cft->root,
+		stripe_filler = find_config_str(cmd->cft->root,
 						"activation/missing_stripe_filler",
 						DEFAULT_STRIPE_FILLER);
 	}
 	dm->stripe_filler = stripe_filler;
 
-	if (!mirror_region_size) {
-		mirror_region_size = 2 * find_config_int(cft->root,
-							 "activation/mirror_region_size",
-							 DEFAULT_MIRROR_REGION_SIZE);
-	}
-	dm->mirror_region_size = mirror_region_size;
-
 	if (!(dm->vg_name = pool_strdup(dm->mem, vg_name))) {
 		stack;
 		goto bad;
@@ -1070,6 +954,8 @@ struct dev_manager *dev_manager_create(const char *vg_name,
 	list_init(&dm->remove_list);
 	list_init(&dm->suspend_list);
 
+	dm->target_state = NULL;
+
 	return dm;
 
       bad:
@@ -1091,7 +977,7 @@ int dev_manager_info(struct dev_manager *dm, const struct logical_volume *lv,
 	/*
 	 * Build a name for the top layer.
 	 */
-	if (!(name = _build_name(dm->mem, lv->vg->name, lv->name, NULL))) {
+	if (!(name = build_dm_name(dm->mem, lv->vg->name, lv->name, NULL))) {
 		stack;
 		return 0;
 	}
@@ -1116,7 +1002,7 @@ int dev_manager_snapshot_percent(struct dev_manager *dm,
 	/*
 	 * Build a name for the top layer.
 	 */
-	if (!(name = _build_name(dm->mem, lv->vg->name, lv->name, NULL))) {
+	if (!(name = build_dm_name(dm->mem, lv->vg->name, lv->name, NULL))) {
 		stack;
 		return 0;
 	}
@@ -1148,7 +1034,7 @@ int dev_manager_mirror_percent(struct dev_manager *dm,
 	/*
 	 * Build a name for the top layer.
 	 */
-	if (!(name = _build_name(dm->mem, lv->vg->name, lv->name, NULL))) {
+	if (!(name = build_dm_name(dm->mem, lv->vg->name, lv->name, NULL))) {
 		stack;
 		return 0;
 	}
@@ -1212,7 +1098,7 @@ static struct dev_layer *_create_layer(struct dev_manager *dm,
 	char *name, *dlid;
 	struct dev_layer *dl;
 
-	if (!(name = _build_name(dm->mem, lv->vg->name, lv->name, layer))) {
+	if (!(name = build_dm_name(dm->mem, lv->vg->name, lv->name, layer))) {
 		stack;
 		return NULL;
 	}
@@ -1282,15 +1168,13 @@ static int _expand_vanilla(struct dev_manager *dm, struct logical_volume *lv,
 	/* Add dependencies for any LVs that segments refer to */
 	list_iterate(segh, &lv->segments) {
 		seg = list_item(segh, struct lv_segment);
-		if (seg->type != SEG_STRIPED && seg->type != SEG_MIRRORED)
-			continue;
 		for (s = 0; s < seg->area_count; s++) {
 			if (seg->area[s].type != AREA_LV)
 				continue;
 			if (!str_list_add(dm->mem, &dl->pre_create,
 					  _build_dlid(dm->mem,
-						      seg->area[s].u.lv.
-						      lv->lvid.s, NULL))) {
+						      seg->area[s].u.lv.lv->
+						      lvid.s, NULL))) {
 				stack;
 				return 0;
 			}
@@ -1311,7 +1195,7 @@ static int _expand_vanilla(struct dev_manager *dm, struct logical_volume *lv,
 	_clear_flag(dlr, VISIBLE);
 	_clear_flag(dlr, TOPLEVEL);
 	_set_flag(dlr, REMOVE);
-	
+
 	/* add the dependency on the real device */
 	if (!str_list_add(dm->mem, &dl->pre_create,
 			  pool_strdup(dm->mem, dlr->dlid))) {
@@ -1336,6 +1220,9 @@ static int _expand_origin_real(struct dev_manager *dm,
 	_clear_flag(dl, VISIBLE);
 	_clear_flag(dl, TOPLEVEL);
 
+	/* Size changes must take effect before tables using it are reloaded */
+	_set_flag(dl, RESUME_IMMEDIATE);
+
 	real_dlid = dl->dlid;
 
 	if (!(dl = _create_layer(dm, NULL, lv))) {
@@ -1656,11 +1543,11 @@ static int _create_rec(struct dev_manager *dm, struct dev_layer *dl)
 	if (dl->info.exists) {
 		if ((suffix = rindex(dl->dlid, '-')))
 			suffix++;
-		newname = _build_name(dm->mem, dm->vg_name, dl->lv->name,
-				      suffix);
+		newname = build_dm_name(dm->mem, dm->vg_name, dl->lv->name,
+					suffix);
 		if (strcmp(newname, dl->name)) {
 			if (!_suspend_parents(dm, dl) ||
-			    !_suspend(dl) || !_rename(dl, newname)) {
+			    !_suspend(dl) || !_rename(dm, dl, newname)) {
 				stack;
 				return 0;
 			}
@@ -1688,7 +1575,6 @@ static int _create_rec(struct dev_manager *dm, struct dev_layer *dl)
 	return 1;
 }
 
-
 static int _build_all_layers(struct dev_manager *dm, struct volume_group *vg)
 {
 	struct list *lvh;
@@ -1756,8 +1642,7 @@ static int _populate_pre_suspend_lists(struct dev_manager *dm)
 				continue;
 			}
 
-			if (!str_list_add(dm->mem, &dep->pre_create,
-					  dl->dlid)) {
+			if (!str_list_add(dm->mem, &dep->pre_create, dl->dlid)) {
 				stack;
 				return 0;
 			}
@@ -1773,8 +1658,7 @@ static int _populate_pre_suspend_lists(struct dev_manager *dm)
 				continue;
 			}
 
-			if (!str_list_add(dm->mem, &dep->pre_suspend,
-					  dl->dlid)) {
+			if (!str_list_add(dm->mem, &dep->pre_suspend, dl->dlid)) {
 				stack;
 				return 0;
 			}
@@ -2086,6 +1970,73 @@ static int _remove_suspended_lvs(struct dev_manager *dm,
 	return 1;
 }
 
+static int _targets_present(struct dev_manager *dm, struct list *lvs)
+{
+	struct logical_volume *lv;
+	struct list *lvh, *segh;
+	struct segment_type *segtype;
+	struct lv_segment *seg;
+	int snapshots = 0, mirrors = 0;
+
+	list_iterate(lvh, lvs) {
+		lv = list_item(lvh, struct lv_list)->lv;
+
+		if (!snapshots)
+			if (lv_is_cow(lv) || lv_is_origin(lv))
+				snapshots = 1;
+
+		if (!mirrors)
+			if (lv->status & PVMOVE)
+				mirrors = 1;
+
+		if (lv->status & VIRTUAL) {
+			list_iterate(segh, &lv->segments) {
+				seg = list_item(segh, struct lv_segment);
+				if (seg->segtype->ops->target_present &&
+				    !seg->segtype->ops->target_present()) {
+					log_error("Can't expand LV: %s target "
+						  "support missing "
+						  "from kernel?",
+						  seg->segtype->name);
+					return 0;
+				}
+			}
+		}
+	}
+
+	if (mirrors) {
+		if (!(segtype = get_segtype_from_string(dm->cmd, "mirror"))) {
+			log_error("Can't expand LV: Mirror support "
+				  "missing from tools?");
+			return 0;
+		}
+
+		if (!segtype->ops->target_present ||
+		    !segtype->ops->target_present()) {
+			log_error("Can't expand LV: Mirror support missing "
+				  "from kernel?");
+			return 0;
+		}
+	}
+
+	if (snapshots) {
+		if (!(segtype = get_segtype_from_string(dm->cmd, "snapshot"))) {
+			log_error("Can't expand LV: Snapshot support "
+				  "missing from tools?");
+			return 0;
+		}
+
+		if (!segtype->ops->target_present ||
+		    !segtype->ops->target_present()) {
+			log_error("Can't expand LV: Snapshot support missing "
+				  "from kernel?");
+			return 0;
+		}
+	}
+
+	return 1;
+}
+
 static int _fill_in_active_list(struct dev_manager *dm, struct volume_group *vg)
 {
 	char *dlid;
@@ -2168,6 +2119,12 @@ static int _action(struct dev_manager *dm, struct logical_volume *lv,
 		}
 	}
 
+	if (!_targets_present(dm, &dm->active_list) ||
+	    !_targets_present(dm, &dm->reload_list)) {
+		stack;
+		return 0;
+	}
+
 	if (!_execute(dm, lv->vg)) {
 		stack;
 		return 0;
@@ -2195,8 +2152,8 @@ int dev_manager_lv_mknodes(const struct logical_volume *lv)
 {
 	char *name;
 
-	if (!(name = _build_name(lv->vg->cmd->mem, lv->vg->name,
-				 lv->name, NULL))) {
+	if (!(name = build_dm_name(lv->vg->cmd->mem, lv->vg->name,
+				   lv->name, NULL))) {
 		stack;
 		return 0;
 	}
