@@ -149,8 +149,8 @@ int insert_pvmove_mirrors(struct cmd_context *cmd,
 				       	seg->area[s].u.pv.pe,
 				       	PVMOVE, allocatable_pvs,
 				       	lv->alloc)) {
-					log_error("Allocation for temporary "
-					  	"pvmove LV failed");
+					log_error("Unable to allocate "
+						  "temporary LV for pvmove.");
 					return 0;
 				}
 				seg->area[s].type = AREA_LV;
